

#include <stdio.h>
#include "mpi.h"

/*  This code performs NITER iterations of an NxN life board.  */

#define N          64
#define NITER      50


void host(int, int) ;
void node(int, int, int) ;

int main(int argc, char* argv []) {
  int size, rank ;

  MPI_Init(&argc, &argv) ;

  MPI_Comm_size(MPI_COMM_WORLD, &size) ;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank) ;

  if(rank == 0)
    host(1, size - 1) ;
  else
    node(1, size - 1, rank - 1) ;

  MPI_Finalize() ;
}

void host(int nodes_base, int nnodes) {
  int blockSizeMax = (N + nnodes - 1) / nnodes ;
  int* block ;
  int iter ;

  /* Handle output */

  block = (int*) malloc(blockSizeMax * sizeof(int [N])) ;

  for(iter = 0 ; iter <= NITER ; iter++) {

    /* Copy current state of board from nodes to a `.pgm' file. */

    char fname [256] ;
    FILE* f ;
    int node_id ;
 
    sprintf(fname, "life%2.2d.pgm", iter) ;

    f = fopen(fname, "w") ;
    fprintf(f, "P2\n%3d  %3d\n%3d\n", N, N, 1) ;

    for(node_id = 0 ; node_id < nnodes ; node_id++) {
      MPI_Status status ;
      int blockLen ;
      int i ;

      MPI_Recv(block, blockSizeMax * N, MPI_INT, nodes_base + node_id,
               1, MPI_COMM_WORLD, &status) ;

      MPI_Get_count(&status, MPI_INT, &blockLen) ;
    
      for(i = 0 ; i < blockLen ; i++) {
        fprintf(f, "%d ", block [i]) ;
        if(i % 20 == 19) fprintf(f, "\n") ;
      }
    }

    fclose(f) ;
  }
}


void node(int nodes_base, int nnodes, int node_id) {

  int blockSizeMax, blockBase, blockSize ;
  int (*block) [N] ;
  int (*neighbours) [N] ;

  MPI_Status status ;

  int i, y ;
  int iter ;

  int node_prev = (node_id - 1 + nnodes) % nnodes ;
  int node_next = (node_id + 1) % nnodes ;

  blockSizeMax = (N + nnodes - 1) / nnodes ;
  blockBase     = blockSizeMax * node_id ;

  if(blockBase + blockSizeMax > N)
    blockSize = N - blockBase ;
  else
    blockSize = blockSizeMax ;

  /* `block' has `blockSize + 2' columns and is offset so that
     subscripts start from -1.  This allows for ghost cells.  */

  block = (int (*) [N]) malloc((blockSize + 2) * sizeof(int [N])) + 1 ;

  /*  Initialise board.  */

  for(i = 0 ; i < blockSize ; i++) {
    for(y = 0 ; y < N ; y++) {
      int x = blockBase + i ;
      if(x == N / 2 || y == N / 2)
        block [i] [y] = 1 ;
      else
        block [i] [y] = 0 ;
    }
  }

  /*  Dump initial state of board to host */

  MPI_Send(block, blockSize * N, MPI_INT, 0, 1, MPI_COMM_WORLD) ;

  /*  Main update loop.  */

  neighbours = (int (*) [N]) malloc(blockSize * sizeof(int [N])) ;

  for(iter = 0 ; iter < NITER ; iter++) {

    /*  Shift this block's upper edge into next neighbour's lower ghost edge */

    MPI_Sendrecv(block [blockSize - 1], N, MPI_INT, nodes_base + node_next, 0,
                 block [-1],            N, MPI_INT, nodes_base + node_prev, 0,
                 MPI_COMM_WORLD, &status) ;

    /*  Shift this block's lower edge into prev neighbour's upper ghost edge */

    MPI_Sendrecv(block [0],         N, MPI_INT, nodes_base + node_prev, 0,
                 block [blockSize], N, MPI_INT, nodes_base + node_next, 0,
                 MPI_COMM_WORLD, &status) ;

    /*  Calculate a block of neighbour sums.  */

    for(i = 0 ; i < blockSize ; i++) {
      for(y = 0 ; y < N ; y++) {
        int y_n = (y - 1 + N) % N ;
        int y_p = (y + 1) % N ;

        neighbours [i] [y] =
             block [i - 1] [y_n] + block [i - 1] [y] + block [i - 1] [y_p] +
             block [i]     [y_n] +                     block [i]     [y_p] +
             block [i + 1] [y_n] + block [i + 1] [y] + block [i + 1] [y_p] ;
      }
    }

    /*  Update block of board values.  */

    for(i = 0 ; i < blockSize ; i++) {
      for(y = 0 ; y < N ; y++) {
        int neighbour = neighbours [i] [y] ;

        if(neighbour < 2 || neighbour > 3)
          block [i] [y] = 0 ;
        if(neighbour == 3)
          block [i] [y] = 1 ;
      }
    }

    /*  Dump current state of board to host */

    MPI_Send(block, blockSize * N, MPI_INT, 0, 1, MPI_COMM_WORLD) ;
  }
}

